#include "waveform.h"
#include "waveitem.c"
#include "faults.c"

#ifndef waveform_c
#define waveform_c

Waveform::Waveform () {
  top = NULL;
  wave_length = 0;
  num_inputs = 0;
  has_ommit = false;
}

Waveform::~Waveform () {
  delete top; // this should delete all the others as well
  top = NULL;
}

// function to set the number of inputs 
// - perhaps called automatically by load 
void Waveform::set_num_inputs(int num_inputs_in) {
  Waveitem *current;  // local for pointing & moving
  string temp_data;
  num_inputs = num_inputs_in;
  // make sure each time-element matches this size
  current = top;
  while (current != NULL) {
    temp_data = current->data;
    if (temp_data.length() != num_inputs) {
      if (temp_data.length() < num_inputs) {
	current->data.append(num_inputs-temp_data.length(), '0');
      } else if (temp_data.length() > num_inputs) {
	current->data.erase(num_inputs);
      }
    }
    current = current->next;
  }
}

// the current length of the waveform 
// - updates the local variable wave_length as well
int Waveform::length() {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  current = top;
  while (current != NULL) {
    i++;
    current = current->next;
  }
  return (wave_length = i);
}

int Waveform::length_ommit() {
  Waveitem *current; // local for pointing & moving
  int i=0;
  current=top;
  if (has_ommit) {
    while (current != NULL) {
      if (!current->ommit) {
	i++;
      }
      current = current->next;
    }
    return i;
  } else {
    return length();
  }
}

// function to set the has_ommit value
void Waveform::ommit(bool yesno) {
  has_ommit = yesno;
}

// function to read the has_ommit value
bool Waveform::ommit() {
  return has_ommit;
}

// returns a value of an input at a specific time
int Waveform::val(int input, int time) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  int ret_val = -1;
  string data;
  // check if time is outside bounds
  if (time < 0 || time >= wave_length) {
    return ret_val;
  }
  current = top;
  while (current != NULL && i < time) {
    current = current->next;
    i++;
  }
  if (current != NULL) {
    data = current->data;
    if (input <= data.length()) {
      ret_val = int(char(data.at(input))-'0');
    }
  }
  return ret_val;
}

string Waveform::val_data(int time) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  string return_val = "";
  if (time < 0 || time >= wave_length) {
    return return_val;
  }
  current = top;
  while (current != NULL && i < time) {
    current = current->next;
    i++;
  }
  if (current != NULL) {
    return_val = current->data;
  }
  return return_val;
}

// inserts a time
int Waveform::add_time(int time) {
  Waveitem *current;  // local for pointing & moving
  string new_data = "";
  new_data.append(num_inputs, '0');
  if (time == 0) {
    // adding to beginning
    if (top == NULL) {
      // if list is empty, then just add
      top = new Waveitem(new_data);
    } else {
      current = top;
      top = new Waveitem(new_data);
      top->next = current;
      current->prev = top;  // point back
    }
  } else if (time >= wave_length) {
    // if time beyond length of list,
    // then it's the same as "padd"
    padd(time);
  } else { // else somewhere in the middle
    int i = 0;
    current = top;
    // iterate through:
    while (current != NULL) {
      i++;
      // if we are at the time:
      if (i == time) {
	// if this is the end of the list
	if (current->next == NULL) {
	  // then at the end of the list
	  current->next = new Waveitem(new_data);
	  current->next->prev = current; // point back
	} else {
	  // adding in the middle:
	  // 1) save next item
	  // 2) add in item
	  // 3) point new to next...
	  Waveitem * next;
	  next = current->next;  // 1)
	  current->next = new Waveitem(new_data);  // 2) [new one is current->next]
	  current->next->prev = current;  // new one points back to previous one
	  current->next->next = next;  // 3)
	  next->prev = current->next;  // next one point back to new one
	}
	// this jumps out of the while
	current = NULL;
      } else {
	// keep moving through the while loop
	current = current->next;
      }
    }
  }
  return length();
}

// adds time to the end 
int Waveform::add_time_end() {
  return add_time(wave_length);
}

// changes a value 
void Waveform::change_val(int input, int time, int value) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  // first, make sure the waveform is long enough
  padd(time);
  current = top;
  while (current != NULL) {
    if (i == time) {
      current->data.replace(input, 1, 1, value?'1':'0');
      current = NULL; // to get out of the loop
    } else {
      current = current->next;
      i++;
    }
  }
}

// change waveitem
void Waveform::change_waveitem(int time, string new_data, bool new_ommit=false) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  // first, make sure the waveform is long enough
  padd(time);
  current = top;
  while (current != NULL) {
    if (i == time) {
      current->data = new_data;
      current->ommit = has_ommit?new_ommit:false;
      current->future_ommit = current->ommit;
      current = NULL;  // to get out of loop
    } else {
      current = current->next;
      i++;
    }
  }
}

// deletes a time
int Waveform::delete_time(int time) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  bool did_delete = false;
  Waveitem *prev, *next;
  current = top;
  if (current != NULL) {
    if (current->next == NULL) {
      // if there is only one in the list
      delete current;
      current = top = NULL;
    } else if (time == 0) {
      // if deleting the top item
      current = top;
      top = top->next;
      top->prev = NULL;  // erase top's prev pointer
      current->next = NULL;  // erase this so delete doesn't propagate
      delete current;
      current = top;
    } else {
      // move on...
      i++;
      prev = top;
      current = top->next;
      while (current != NULL && i <= time) {
	if (i == time && !did_delete) {
	  // save where I point to
	  next = current->next;
	  // delete me
	  current->next = NULL;  // erase this so delete doesn't propagate
	  delete current;
	  // set me to previous
	  current = prev;
	  // set next to saved
	  current->next = next;
	  next->prev = current;  // point back
	  // did delete
	  did_delete = true;
	}
	// save "previous" (next iteration's previous)
	prev = current;
	// move to next one
	current = current->next;
	// increment time:
	i++;
      }
    }
  }
  // update length
  return length();
}

// pads a waveform out to a certain time length
int Waveform::padd(int time) {
  string temp_string = "";
  return padd_internal(time, temp_string.append(num_inputs, '0'));
}

// pads a waveform out to a certain time length
int Waveform::padd_same(int time) {
  return padd_internal(time, val_data(wave_length-1));
}

// pads a waveform out to a certain time length
int Waveform::padd_internal(int time, string new_data) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  if (time >= wave_length) {
    if (top == NULL) {
      top = new Waveitem(new_data);
    }
    current = top;
    while(current->next != NULL) {
      current = current->next;
      i++;
    }
    while (i < time) {
      current->next = new Waveitem(new_data);
      current->next->prev = current;  // point back
      current = current->next;
      i++;
    }
  }
  return length();
}

// wave functions
void Waveform::append_wave(Waveform *w) {
  insert_wave(w, length());
}

void Waveform::insert_wave(Waveform *w, int time) {
  Waveitem *current;    // local for pointing & moving
  int i;
  i = time;
  padd(time - 1);
  current = w->top;
  while (current != NULL) {
    add_time(i);
    change_waveitem(i, current->data, current->ommit);
    current = current->next;
    i++;
  }
}

// erases, and creates new wave that's random
void Waveform::rand_wave_new(int new_wave_length, int new_num_inputs) {
  delete top;
  top = NULL;
  num_inputs = new_num_inputs;
  rand_wave_time(0, new_wave_length);
}

// randomize all values of the wave 
void Waveform::rand_wave() {
  rand_wave_time(0, wave_length);
}

void Waveform::rand_wave_time(int start_time, int rand_length) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  // make sure the right length:
  padd(start_time + rand_length - 1);
  current = top;
  while (current != NULL) {
    if (i >= start_time + rand_length) {
      return;
    } else if (i >= start_time) {
      current->data = rand_data();
    }
    current = current->next;
    i++;
  }
}

string Waveform::rand_data () {
  string new_data = "";
  int i;
  // create random string:
  for (i=0; i<num_inputs; i++) {
    new_data.append(1, int(2 * rand()/(RAND_MAX+1.0))?'1':'0');
  }
  return new_data;
}

// erases, and creates new wave that's all zeros
void Waveform::zero_wave_new(int new_wave_length, int new_num_inputs) {
  delete top;
  top = NULL;
  num_inputs = new_num_inputs;
  zero_wave_time(0, new_wave_length);
}

// zeros out all values of the wave 
void Waveform::zero_wave() {
  zero_wave_time(0, wave_length);
}

void Waveform::zero_wave_time(int start_time, int zero_length) {
  Waveitem *current;  // local for pointing & moving
  int i = 0;
  // make sure the right length:
  padd(start_time + zero_length - 1);
  current = top;
  while (current != NULL) {
    if (i >= start_time + zero_length) {
      return;
    } else if (i >= start_time) {
      current->data = zero_data();
    }
    current = current->next;
    i++;
  }
}

string Waveform::zero_data() {
  string new_data = "";
  // create blank string (all zeros)
  // casts to "unsigned int" is to overcome compilier ambiguity over insert function
  return new_data.insert((unsigned int) 0, (unsigned int) num_inputs, '0');
}

// fault functions
void Waveform::ommit_all() {
  set_ommits(true);
}

void Waveform::ommit_none() {
  set_ommits(false);
}

void Waveform::set_ommits(bool yesno) {
  Waveitem *current;  // local for pointing & moving
  current = top;
  while (current != NULL) {
    current->ommit = yesno;
    current->future_ommit = yesno;
    current = current->next;
  }
}

void Waveform::ommits_update() {
  Waveitem *current;  // local for pointing & moving
  current = top;
  while (current != NULL) {
    current->ommit = current->future_ommit;
    current = current->next;
  }
}

void Waveform::set_fault_ommits(Faults * f) {
  Waveitem *current; // for pointing at waveitems in waveform
  Faultitem *current_f; // for pointing at fault items in f
  int thistime=-1;
  int i;
  ommit_all(); // initially set all to be ommitted
  current_f = f->top;
  // if not an empty list:
  if (current_f != NULL) {
    thistime = current_f->fault_time;
    // want to set future ommits to false for all before initial time
    // (could call set_future_ommit(...) each time, but this is more efficient
    current = top;
    for (i=0; i<thistime; i++) {
      if (current != NULL) {
	current->future_ommit = false;
	current = current->next;
      }
    }
    // clear "thistime" variable
    thistime = -1;
    while (current_f != NULL) {
      // there may be multiple faults at a time,
      // but (for efficiency) only want to set each time once
      if (current_f->fault_time != thistime) {
	thistime = current_f->fault_time;
	// set future ommit checks to make sure "thistime"
	// isn't past the end of the waveform
	set_future_ommit(thistime, false);
      }
      current_f = current_f->next;
    }
  }
  // after setting future ommits, update so ommits match:
  ommits_update();
}

void Waveform::update_fault_ommits(Faults *f) {
  Waveitem *current; // for pointing at waveitems in waveform
  Waveitem *back; // for moving back in list
  Faultitem *current_f; // for pointing at fault items in f
  int current_time = -1; // the time current points at
  int thistime=-1;
  bool matched=false;
  int i;
  current_f = f->top;
  current = top;
  if (current_f != NULL && current != NULL) {
    // if not an empty fault list and there is a waveform:
    while (current_f != NULL) {
      if (thistime != current_f->fault_time) {
	thistime = current_f->fault_time;
	// set current to be pointing at thistime
	current = top;
	current_time = -1;
	while (current != NULL && current_time < thistime) {
	  current_time++;
	  // if we haven't found the time yet AND
	  // the next one isn't null (which would fall off the end of the list)
	  // then move forward
	  if (current_time != thistime && current->next != NULL) {
	    current = current->next;
	  }
	}
	// now: current points to current_time which equals thistime
	back = current; // don't want to mess with current, so use back
	while (back != NULL && !back->ommit) {
	  back = back->prev; // move backwards through the list
	}
	// now either back points nothing (went off top of list)
	// or points at an waveform element before current fault 
	// that has been ommited
	if (back != NULL) {
	  // set this waveform element to be ommitted in the future:
	  back->future_ommit = false;
	}
      }
      // already set the ommit for "thistime" so move just move one
      current_f = current_f->next;
    } // end while loop
    // went through all faults setting future ommit
    // now update real ommits (to be future ommits)
    ommits_update();
  }
}

void Waveform::set_future_ommit(int time, bool yesno) {
  Waveitem *current; // for pointing at waveitems in waveform
  int i=0;
  int temp_length;
  int temp_string;
  current = top;
  if (current != NULL) {
    // thought at first I should
    // padd out to time duplicating last
    // >> padd_same(time); 
    // but now it should not go past the end of the
    // wavelength ... if so, just set to last
    if (time >= wave_length) {
      time = wave_length-1;
    }
    while (current != NULL) {
      if (i == time) {
	current->future_ommit = yesno;
	current = NULL; // break out of loop
      } else {
	i++;
	current = current->next;
      }
    } // end while loop
  }
}

void Waveform::print() {
  Waveitem *current;  // local for pointing & moving
  int i=0;
  current = top;
  while (current != NULL) {
    cout << i;
    if (has_ommit) {
      cout << " (" << (current->ommit ? 1 : 0);
      if (current->ommit != current->future_ommit) {
	cout << "/" << (current->future_ommit ? 1 : 0);
      }
      cout << ")";
    }
    cout << ": " << current->data << "\n";
    current = current->next;
    i++;
  }
}

// load the waveform from a file
int Waveform::load(string file_name) {
  // first delete any possible current list:
  delete top;
  top = NULL;
  char wave[100];
  bool keep_parsing = true;
  int time = 0;
  fstream file;
  file.open(file_name.c_str(), ios::in);
  while (!file.eof () && keep_parsing) {
    file.getline(wave, 100, '\n');
    keep_parsing = parse_wave((string) wave);
  }
  file.close();
  if (top != NULL) {
    set_num_inputs(top->data.length());
  }
  return length();
}

bool Waveform::parse_wave(string wave) {
  Waveitem *current;  // local for pointing & moving
  if (wave.length() > 0) {
    if (top == NULL) {
      if (has_ommit) {
	top = new Waveitem(wave.substr(1));
	top->ommit = (wave.at(0)=='1');
	top->future_ommit = top->ommit;
      } else {
	top = new Waveitem(wave);
	top->ommit = false;
	top->future_ommit = false;
      }
      current = top;
    } else {
      current = top;
      // get to end of the list
      while (current->next != NULL) {
	current = current->next;
      }
      if (has_ommit) {
	current->next = new Waveitem(wave.substr(1));
	current->next->ommit = (wave.at(0)=='1');
	current->next->future_ommit = current->next->ommit;
      } else {
	current->next = new Waveitem(wave);
      }
      current->next->prev = current; // point back
    }
    return true;
  } else {
    return false;
  }
}

// loads an array into the waveform (overwriting) 
void Waveform::load_array(float * arr, int arr_size, int input_num, int start_time) {
  int i=0;
  // make sure the right length:
  padd(start_time + arr_size - 1);
  for (i=0; i<arr_size; i++) {
    change_val(input_num, start_time+i, (arr[i]>0)?1:0);
  }
}

// save the waveform to a file
// including ommit information (if has_ommit flag is set)
int Waveform::save_ommit(string file_name) {
  Waveitem *current;  // local for pointing & moving
  fstream file;
  file.open(file_name.c_str(), ios::out);
  if (file.is_open()) {
    current = top;
    while (current != NULL) {
      if (has_ommit) {
	file << current->ommit;
      }
      file << current->data << "\n";
      current = current->next;
    }
    file.close();
  } 
  return length();
}

// save the waveform (ommitting vectors if has_ommit flag is set)
int Waveform::save(string file_name) {
  Waveitem *current;  // local for pointing & moving
  int i=0;
  fstream file;
  if (has_ommit) {
    file.open(file_name.c_str(), ios::out);
    if (top != NULL) {
      current = top;
      while (current != NULL) {
	if (! current->ommit) {
	  i++;
	  file << current->data << "\n";
	}
	current = current->next;
      }
    }
    file.close();
    return i;
  } else {
    return save_ommit(file_name);
  }
}

// saves waveform as .dwe format:
int Waveform::save_dwe(string file_name) {
  return save_dwe(file_name, false); // pass along call
}

int Waveform::save_dwe(string file_name, bool needclock) {
  Waveitem *current;  // local for pointing & moving
  string last_values;
  int time=0;
  int period=10;
  int input;
  int input_offset; // due to presense of clock or not
  fstream file;
  file.open(file_name.c_str(), ios::out);
  if (top != NULL) {
    current = top;
    file << "0\n"; // zero fault observation points
    if (needclock) {
      file << "1\n"; // clock count
      file << time << " " << period << " 5 0\n"; // index, period, rise, fall
      input_offset = 1;
    } else {
      file << "0\n"; // no clock
      input_offset = 0;
    }
    // initial conditions: "FORCE input# bit# value"
    for (input=0; input<num_inputs; input++) {
      file << "2 " << (input+input_offset) << " 0 " << current->data.at(input) << "\n";
    }
    last_values = current->data;
    current = current->next;
    while (current != NULL) {
      if (current->data == "END") { // sometimes .dgp have END at the end
	current = NULL;
      } else if (current->data.length() > 0) {
	time += period;
	file << "5\n"; // STROBE
	file << "1 " << time << "\n"; // next time
	for (input=0; input<num_inputs; input++) {
	  if (current->data.at(input) != last_values.at(input)) {
	    file << "3 " << (input+input_offset) << " 0 " << current->data.at(input) << "\n";
	  }
	}
	last_values = current->data;
	current = current->next;
      } else {
	current = NULL; // bail out
      }
    } // end while
    // run the last waveform for a couple times:
    file << "5\n"; // strobe
    file << "1 " << (time+=period) << "\n";
    file << "5\n"; // strobe
    file << "1 " << (time+=period) << "\n";
    file << "0\n";
  }
  file.close();
  return time;
}

// saves waveform as .dwe format:
int Waveform::save_dwe_TUFTsim(string file_name, int period) {
  return save_dwe_TUFTsim(file_name, period, false);
}

int Waveform::save_dwe_TUFTsim(string file_name, int period, bool needclock) {
  Waveitem *current;  // local for pointing & moving
  string last_values;
  int time=0;
  if (!(period > 0)) {
    period = 100;
  }
  int input;
  fstream file;
  file.open(file_name.c_str(), ios::out);
  if (top != NULL) {
    current = top;
    file << "0\n"; // zero fault observation points
    file << "0\n"; // second zero for TUFTsim
    //    file << "1\n"; // clock count
    //    file << time << " " << period << " 5 0\n"; // index, period, rise, fall

    // initial conditions: "FORCE input# bit# value"
    for (input=0; input<num_inputs; input++) {
      file << "2 " << (input) << " 0 " << current->data.at(input) << "\n";
    }
    last_values = current->data;
    current = current->next;
    while (current != NULL) {
      if (current->data == "END") { // sometimes .dgp have END at the end
	current = NULL;
      } else if (current->data.length() > 0) {
	time += period;
	file << "5\n"; // STROBE
	file << "1 " << time << "\n"; // next time
	for (input=0; input<num_inputs; input++) {
	  if (current->data.at(input) != last_values.at(input)) {
	    file << "3 " << (input) << " 0 " << current->data.at(input) << "\n";
	  }
	}
	last_values = current->data;
	current = current->next;
      } else {
	current = NULL; // bail out
      }
    } // end while
    // run the last waveform for a couple times:
    file << "5\n"; // strobe
    file << "1 " << (time+=period) << "\n";
    file << "5\n"; // strobe
    file << "1 " << (time+=period) << "\n";
    file << "0\n";
  }
  file.close();
  return time;
}

// saves waveform to an array
void Waveform::save_array(float * arr, int arr_size, int input_num, int start_time) {
  Waveitem *current;  // local for pointing & moving
  int i=0;
  // first clear the array
  for (i=0; i<arr_size; i++) {
    arr[i] = 0;
  }
  for (i=0; i<arr_size; i++) {
    if (i+start_time < wave_length) {
      arr[i] = val(input_num, i+start_time);
    }
  }
}

#endif
